package assemble;

import java.util.ArrayList;
import java.util.HashMap;

import shared.Tools;
import structures.ByteBuilder;

public class BubblePopper {
	
	BubblePopper(ArrayList<Contig> allContigs_, HashMap<Integer, ArrayList<Edge>> destMap_, int kbig_){
		allContigs=allContigs_;
		destMap=destMap_;
		kbig=kbig_;
		minLen=2*kbig-1;
	}
	
	int expand(Contig c) {
//		if(true) {return 0;}
		if(verbose){System.err.println("\n\n*expand: "+c.name()+"\n");}
		assert(!c.used);
		center=c;
		int count=0;
		if(center.rightForwardBranch()){
			while(expandRight()){
				count++;
			}
		}
		if(center.leftForwardBranch()){
			center.flip(destMap.get(center.id));
			while(expandRight()) {
				count++;
			}
			center.flip(destMap.get(center.id));
		}
		return count;
	}
	
	private boolean expandRight(){
		//Reset shared state
		dest=null;
		lastMutualDest=-1;
		lastMutualDestOrientation=-1;
		
		if(verbose){System.err.println("expandRight: "+center.name());}
		
		if(!center.rightForwardBranch() || center.rightEdgeCount()<1){
			if(verbose){System.err.println("Returned because not forward branch or no edges.");}
			return false;
		}
		ArrayList<Edge> outbound=center.rightEdges;
		final Edge leftMidEdge=findRepresentativeMidEdge(outbound);
		
		if(leftMidEdge==null){
			if(verbose){System.err.println("No leftMidEdge.");}
			return false;
		}
		final Contig mid=allContigs.get(leftMidEdge.destination);
		if(mid==null || mid.length()<minLen){
			if(verbose){System.err.println("No mid, or mid too short ("+(mid==null ? 0 : mid.length())+").");}
			return false;
		}
		
		if(verbose){System.err.println("\nFinding mutualDest for center node.");}
		final int mutualDest=findMutualDest(outbound);
		final int mutualDestOrientation=lastMutualDestOrientation;
		final boolean mutualDestRight=((lastMutualDestOrientation&2)==2);
		if(verbose){System.err.println("mutualDest="+mutualDest+", mutualDestOrientation="+mutualDestOrientation+", mutualDestRight="+mutualDestRight);}
		
		if(mutualDest<0 || mutualDestOrientation<0){
			if(verbose){System.err.println("Bad mutual destination.");}
			return false;
		}
		//At this point we are fairly confident everything is safe, but still need to run more tests.
		
		dest=allContigs.get(mutualDest);
		if(dest.used){return false;}
//		assert(!dest.used);//This happened; not sure how
		if(dest.id==center.id){
			if(verbose){System.err.println("Self mutual destination.");}
			return false;
		}
		
		if(mutualDestRight && !dest.rightForwardBranch()){
			if(verbose){System.err.println("Mutual destination does not have a right F-branch: "+dest.name());}
			return false;
		}
		if(!mutualDestRight && !dest.leftForwardBranch()){
			if(verbose){System.err.println("Mutual destination does not have a left F-branch: "+dest.name());}
			return false;
		}
		final ArrayList<Edge> destOutbound=mutualDestRight ? dest.rightEdges : dest.leftEdges;
		if(destOutbound==null){
			if(verbose){System.err.println("No dest outbound edges.");}
			return false;
		}

		if(verbose){System.err.println("\nFinding mutualDest for right node.");}
		final int mutualDest2=findMutualDest(destOutbound);
		if(mutualDest2<0 || mutualDest2!=center.id){
			if(verbose){System.err.println("MutualDest2 is not the correct origin: "+center.id+"!="+mutualDest2+"; "+dest.name());}
			return false;
		}
		
		ArrayList<Contig> midNodes=fetchMidNodes(outbound, true);
		//Now we have all intermediate nodes, which are flipped into the correct orientation.
		
		if(!midNodesConcur(midNodes)){
			if(verbose){System.err.println("Mid nodes do not concur.");}
			return false;
		}
		
		//At this point all tests have been run and we are confident that this is a simple bubble.
		if(mutualDestRight){
			if(verbose){System.err.println("Flipping dest node.");}
			dest.flip(destMap.get(dest.id));
		}
		
		//Now the destination is also flipped into the correct orientation.
		final Edge rightMidEdge=mid.getRightEdge(dest.id, 1);
		if(rightMidEdge==null){return false;}

		if(verbose){System.err.println("Popping bubble between "+center.id+" and "+dest.id);}
		return pop(center, dest, mid, leftMidEdge, rightMidEdge, midNodes);
		
	}
	
	private boolean pop(Contig left, Contig right, Contig mid, Edge leftMidEdge, Edge rightMidEdge, ArrayList<Contig> midNodes){
		bb.clear();
		final int originalLeftLength=left.length();
		
		//Append path
		bb.append(left.bases);
		if(leftMidEdge.bases!=null){
			for(int i=0; i<leftMidEdge.bases.length-1; i++){
				bb.append(leftMidEdge.bases[i]);
			}
		}
		for(int i=kbig-1, lim=mid.length()-kbig+1; i<lim; i++){
			bb.append(mid.bases[i]);
		}
		if(rightMidEdge.bases!=null){
			for(int i=0; i<rightMidEdge.bases.length-1; i++){
				bb.append(rightMidEdge.bases[i]);
			}
		}
		bb.append(right.bases);
		left.bases=bb.toBytes();
		
		//Cleanup
		right.used=true;
		mid.used=true;
		for(Contig c : midNodes){
//			c.used=true;
			destMap.remove(c.id);
			c.leftEdges=null;
			c.rightEdges=null;
			//Now an associate contig.
		}
		left.rightEdges.clear();
		if(right.rightEdgeCount()>0){
			for(Edge e : right.rightEdges){
				e.origin=left.id;
				left.addRightEdge(e);
			}
		}else{left.rightEdges=null;}
		
		ArrayList<Edge> inbound=destMap.get(right.id);
		if(inbound!=null){
			for(Edge e : inbound){
				e.destination=left.id;
			}
		}

		left.maxCov=Tools.max(left.maxCov, right.maxCov, mid.maxCov);
		left.minCov=Tools.min(left.maxCov, right.maxCov);
		left.rightCode=right.rightCode;
		left.rightRatio=right.rightRatio;
		double coverageSum=left.coverage*(originalLeftLength-kbig+1)+right.coverage*(right.length()-kbig+1);
		left.coverage=(float)(coverageSum/(left.length()-kbig+1));
		
		expansions++;
		contigsAbsorbed+=(1+midNodes.size());
		
		if(verbose){
			System.err.println("*Result: "+center.name());
		}
		
		return true;
	}
	
	private Edge findRepresentativeMidEdge(ArrayList<Edge> edges){
		Edge midEdge=null;
		Contig mid=null;
		for(Edge e : edges){
			Contig c=allContigs.get(e.destination);
			if(midEdge==null){
				midEdge=e;
				mid=c;
			}else{
				if(mid.length()<minLen && c.length()>=minLen){
					midEdge=e;
					mid=c;
				}else if(c.length()>=minLen) {
					if(e.depth>midEdge.depth || (e.depth==midEdge.depth && c.length()>mid.length())) {
						midEdge=e;
						mid=c;
					}
				}
			}
		}
		return midEdge;
	}
	
	private boolean midNodesConcur(ArrayList<Contig> midNodes){
		int leftDest=-1;
		int rightDest=-1;
		for(Contig c : midNodes){
			if(c.leftEdges==null){
				if(verbose){System.err.println("No midnode left edges for "+c.name());}
				return false;
			}
			if(c.rightEdges==null){
				if(verbose){System.err.println("No midnode right edges for "+c.name());}
				return false;
			}
			for(Edge e : c.leftEdges){
				if(leftDest<0){leftDest=e.destination;}
				else if(leftDest!=e.destination){
					if(verbose){System.err.println("Different left destination: "+leftDest+" vs "+e.destination);}
					return false;
				}
				if(e.origin==e.destination){
					if(verbose){System.err.println("Left midnode loop for "+c.name());}
					return false;
				}
			}
			for(Edge e : c.rightEdges){
				if(rightDest<0){rightDest=e.destination;}
				else if(rightDest!=e.destination){
					if(verbose){System.err.println("Different right destination: "+rightDest+" vs "+e.destination);}
					return false;
				}
				if(e.origin==e.destination){
					if(verbose){System.err.println("Right midnode loop for "+c.name());}
					return false;
				}
			}
			ArrayList<Edge> incoming=destMap.get(c.id);
			if(incoming==null){
				if(verbose){System.err.println("No incoming edges.");}
				assert(false);
				return false;
			}
			for(Edge e : incoming){
				if(e.origin!=center.id && e.origin!=dest.id){
					if(verbose){System.err.println("Midnode incoming loop for "+c.name());}
					return false;
				}
			}
		}
		if(verbose){System.err.println("Mid nodes concur.");}
		assert(leftDest<0 || leftDest==center.id);
		assert(rightDest<0 || rightDest==dest.id);
		return leftDest>=0 && rightDest>=0;
	}
	
	private ArrayList<Contig> fetchMidNodes(ArrayList<Edge> outbound, boolean flipAsNeeded){
		ArrayList<Contig> midNodes=new ArrayList<Contig>(outbound.size());
		for(Edge e : outbound){
			Contig mid=allContigs.get(e.destination);
			assert(!mid.used);
			if(!mid.used){
				midNodes.add(mid);
				if(flipAsNeeded && e.destRight()){
					mid.flip(destMap.get(mid.id));
				}
			}
		}
		return midNodes;
	}
	
	private int	findMutualDest(ArrayList<Edge> edges){
		if(verbose){System.err.println("findMutualDest("+edges+")");}
		lastMutualDest=-2;
		lastMutualDestOrientation=-1;
		for(Edge e : edges){
			if(verbose){System.err.println("\nConsidering inbound edge "+e);}
			Contig mid=allContigs.get(e.destination);
			if(verbose){System.err.println("Considering mid node "+mid.name());}
			if(mid==center){
				if(verbose){System.err.println("Mid node is center.");}
				return -1;
			}
			ArrayList<Edge> outbound=(e.destRight() ? mid.leftEdges : mid.rightEdges);
			if(verbose){System.err.println("e.destRight()="+e.destRight()+", using "+(outbound==mid.leftEdges ? "mid.leftEdges" : "mid.rightEdges"));}
			if(outbound!=null){
				for(Edge o : outbound){
					if(verbose){System.err.println("Considering mid node edge "+o);}
					if(lastMutualDest<0){
						lastMutualDest=o.destination;
						lastMutualDestOrientation=(o.orientation&2);
						if(verbose){System.err.println("Mutual dest is now "+lastMutualDest+", orientation "+lastMutualDestOrientation);}
					}else if(lastMutualDest!=o.destination){
						if(verbose){System.err.println("Mismatched mutual dest: "+lastMutualDest+" versus "+o.destination);}
						return -1;
					}else if(lastMutualDestOrientation!=(o.orientation&2)){
						if(verbose){System.err.println("Mismatched mutual orientation: "+lastMutualDestOrientation+" versus "+(o.orientation&2));}
						return -1;
					}
				}
			}
		}
		return lastMutualDest;//Can be -2 if there is no destination
	}
	
//	boolean testRightBubble(Contig c, ArrayList<Edge> inbound_0){
//		ArrayList<Edge> outbound=c.rightEdges;
//		int mutualDest=findMutualDest(outbound);
//		if(mutualDest<0){return false;}
//		
//		IntList midContigs=new IntList(4);
//		ArrayList<Edge> inbound=new ArrayList<Edge>(inbound_0.size());
//		for(Edge e : inbound_0){
//			if(e.destRight()){
//				inbound.add(e);
//				midContigs.add(e.destination);
//			}
//		}
//		for(Edge e : outbound){
//			midContigs.add(e.destination);
//		}
//		
//		midContigs.sort();
//		for(int i=0; i<midContigs.size(); i+=2){
//			if(midContigs.size()<i+2 || midContigs.get(i)!=midContigs.get(i+1)){
//				assert(false) : midContigs; //for testing; should be removed
//				return false;
//			}
//		}
//		midContigs.shrinkToUnique();
//		//At this point we have established that all edges are bidirectional.
//		
//		int mutualSource=findMutualSource(inbound);
//		if(mutualSource<0){return false;}
//		int mutualLength=findMutualLength(outbound, 3);
//		if(mutualLength>3*kbig){return false;}
//		if(alternativeSource(outbound)){return false;}
//		
//		if(mutualSource==c.id && mutualDest>c.)
//	}
//	
//	boolean alternativeSource(ArrayList<Edge> edges){
//		int source=-2;
//		for(Edge e : edges){
//			Contig mid=allContigs.get(e.destination);
//			ArrayList<Edge> inbound=(e.destRight() ? mid.rightEdges : mid.leftEdges);
//			if(inbound==null || inbound.size()>1){return false;}
//			if(inbound!=null){
//				for(Edge i : inbound){
//					if(source<0){source=i.destination;}
//					else if(source!=i.destination){return true;}
//				}
//			}
//		}
//		return false;
//	}
//	
//	int	findMutualDest(ArrayList<Edge> edges){
//		int dest=-2;
//		int orientation=-1;
//		for(Edge e : edges){
//			Contig mid=allContigs.get(e.destination);
//			ArrayList<Edge> outbound=(e.destRight() ? mid.leftEdges : mid.rightEdges);
////			ArrayList<Edge> inbound=(e.destRight() ? mid.rightEdges : mid.leftEdges);
////			if(inbound!=null && inbound.size()>2){return -1;}
//			if(outbound!=null){
//				for(Edge o : outbound){
//					if(dest<0){
//						dest=o.destination;
//						orientation=(o.orientation&2);
//					}else if(dest!=o.destination){return -1;}
//					else if(orientation!=(o.orientation&2)){return -1;}
//				}
//			}
//		}
//		return dest;//Can be -2 if there is no destination
//	}
//	
//	int	findMutualSource(ArrayList<Edge> edges){
//		int source=-2;
//		for(Edge e : edges){
//			Contig mid=allContigs.get(e.origin);
//			ArrayList<Edge> outbound=(e.sourceRight() ? mid.leftEdges : mid.rightEdges);
//			if(outbound!=null){
//				for(Edge o : outbound){
//					if(source<0){source=o.origin;}
//					else if(source!=o.origin){return -1;}
//				}
//			}
//		}
//		return source;//Can be -2 if there is no source
//	}
//	
//	int	findMutualLength(ArrayList<Edge> edges, int leeway){
//		int min=Integer.MAX_VALUE;
//		int max=-1;
//		for(Edge e : edges){
//			min=Tools.min(min, e.length);
//			max=Tools.max(max, e.length);
//		}
//		return (max-min>leeway ? -1 : max);
//	}
	
	final ArrayList<Contig> allContigs;
	final HashMap<Integer, ArrayList<Edge>> destMap;
	final int kbig;
	final int minLen;
	final ByteBuilder bb=new ByteBuilder();
	
	Contig center=null;
	Contig dest=null;
	
	int lastMutualDest=-1;
	int lastMutualDestOrientation=-1;
	
	int expansions=0;
	int contigsAbsorbed=0;
	
	static boolean verbose=false;
}
